//
// Benchmarking serialising and deserializing triangular mats of 64-bit ints.
//
// build: javac Bench2Serialize.java
//
// run: time taskset 2 java Bench2Serialize
//

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

public class Bench2Serialize {
    static long[] fibs;

    public static void main(String[] args)
        throws ClassNotFoundException, IOException {
	makeUpto(91);
        System.out.print("[WARMUP]");
        bench(200);  // warm up
	if (args.length == 0)
            for (int d = 500; d <= 4000; d = d + 500)
                bench(d);
	else
	    bench(Math.max(0, (new Integer(args[0])).intValue()));
    }

    public static void makeUpto(int n) {
	fibs = new long[n];
        fibs[0] = 0;
        fibs[1] = 1;
        for (int i = 2; i < n; i++)
	    fibs[i] = fibs[i-1] + fibs[i-2];
    }

    public static long get(int i) {
	return fibs[i];
    }

    public static byte[] serialize(FibMat m)
        throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ObjectOutputStream oout = new ObjectOutputStream(baos);
        oout.writeObject(m);
        oout.close();
        return baos.toByteArray();
    }

    public static FibMat deserialize(byte [] bs)
        throws ClassNotFoundException, IOException {
        ByteArrayInputStream bais = new ByteArrayInputStream(bs);
        ObjectInputStream oin = new ObjectInputStream(bais);
        FibMat m = (FibMat)oin.readObject();
        oin.close();
        return m;
    }

    public static void bench(int dim)
        throws ClassNotFoundException, IOException {
        FibMat m0, m1;
        byte[] bs;
        boolean ok;
        long ts0, ts1;

        m0 = new FibMat(dim);
        System.out.println("dim=" + dim + "  longs=" + (dim+1) * dim / 2);
        System.gc();
        System.out.print("serialising...   ");
        ts0 = System.nanoTime();
        bs = serialize(m0);
        ts1 = System.nanoTime();
        System.out.println((ts1 - ts0) / 1000 + "us");
        System.out.println("bytes=" + bs.length);
        System.gc();
        System.out.print("deserialising... ");
        ts0 = System.nanoTime();
        m1 = deserialize(bs);
        ts1 = System.nanoTime();
        System.out.println((ts1 - ts0) / 1000 + "us");
        System.gc();
        System.out.print("checking...      ");
        ts0 = System.nanoTime();
        ok = m0.equal(m1);
        ts1 = System.nanoTime();
        System.out.println((ts1 - ts0) / 1000 + "us");
        System.out.println("ok=" + ok);
        System.out.println();
        System.gc();
    }
}


class FibVec implements Serializable {
    long[] elt;

    public FibVec(int dim, int sign) {
        this.elt = new long[dim];
        for (int i = 0; i < dim; i++) {
            int j = i % 91;
	    this.elt[i] = sign * Bench2Serialize.get(j);
            if (j % 2 != 0)
		this.elt[i] = - this.elt[i];
        }
    }

    public long[] getElt() {
        return this.elt;
    }

    public boolean equal(FibVec v) {
        if (v == null)
            return false;
        if (this.getElt().length != v.getElt().length)
	    return false;
        boolean eq = true;
        for (int i = 0; i < this.getElt().length; i++)
	    if (eq && this.getElt()[i] != v.getElt()[i])
                eq = false;
        return eq;
    }
}


class FibMat implements Serializable {
    FibVec[] row;

    public FibMat(int dim) {
        this.row = new FibVec[dim];
        for (int i = 0; i < dim; i++) {
            int sign = 1;
            if (sign % 2 == 0)
		sign = -1;
	    this.row[i] = new FibVec(dim, sign);
        }
    }

    public FibVec[] getRows() {
        return this.row;
    }

    public boolean equal(FibMat m) {
        if (m == null)
            return false;
        if (this.getRows().length != m.getRows().length)
	    return false;
        boolean eq = true;
        for (int i = 0; i < this.getRows().length; i++)
	    if (eq && !this.getRows()[i].equal(m.getRows()[i]))
                eq = false;
        return eq;
    }
}
