#lang racket/base

;;
;; Benchmarking serialising and deserializing triangular matrices of fixnums;
;; version using racket's fasl library -- will not run in pycket.
;;

(require racket/fasl)

(define (fib-upto n)
  (let ([f (make-vector n)])
    (for ([i (in-range n)])
      (if (< i 2)
        (vector-set! f i i)
        (vector-set! f i (+ (vector-ref f (- i 1)) (vector-ref f (- i 2))))))
    f))

;; Vector of Fibonacci numbers (from 0 to 90) fitting into 64 bit.
(define fib-upto-91 (fib-upto 91))

(define (fibvector dim [sign 1])
  (for/vector #:length dim
    ([j (in-range dim)])
    (* (if (even? j) sign (- sign)) (vector-ref fib-upto-91 (modulo j 91)))))

(define (fibmatrix dim)
  (for/vector #:length dim
    ([i (in-range dim)])
    (fibvector (add1 i) (if (even? i) 1 -1))))


;; Measure time for serialising, deserialising and checking fibmatrices
(define (bench dim reps)
  (printf "dim=~a  fixnums=~a  repeats=~a\n"
    dim (/ (* dim (add1 dim)) 2) reps)
  (define m0s (for/list ([i (in-range reps)]) (fibmatrix dim)))
  (collect-garbage)
  (printf "serialsing...    ")
  (define bss (time (for/list ([m m0s]) (s-exp->fasl m))))
  (printf "bytes=~a\n" (bytes-length (car bss)))
  (collect-garbage)
  (printf "deserialsing...  ")
  (define m1s (time (for/list ([bs bss]) (fasl->s-exp bs))))
  (collect-garbage)
  (printf "checking...      ")
  (define ok (time (equal? m0s m1s)))
  (printf "ok=~a\n\n" ok)
  (collect-garbage))


;; main
(printf "[WARMUP]")
(bench 200 1)  ;; warm up
(define args (vector->list (current-command-line-arguments)))
(if (null? args)
  (for ([dim (in-range 500 4001 500)])
    (bench dim 1))
  (let ([dim (string->number (car args))]
        [reps  (if (null? (cdr args)) 1 (string->number (cadr args)))])
    (and dim (integer? dim) reps (integer? reps)
      (bench (max 0 dim) (max 1 reps)))))
