#lang racket/base

;;
;; SERIALISATION of acyclic data, with no regard for sharing (64 bit arch)
;;

(require racket/fixnum)
(require racket/flonum)

(provide serialized-size serialize-to! serialize deserialize-from deserialize)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Size estimation of serialized bytestrings

;; (serialized-size x) -> (or/c positive-integer? #f)
;;   x : any/c
;;
;; Return the size in bytes of the serialised representation of `x`,
;; or `#f` if `x` is not serialisable.
;;
(define (serialized-size x)
  (size/aligned x 0))


;; (size/aligned x pos8) -> (or/c positive-integer? #f)
;;   x    : any/c
;;   pos8 : (integer-in 0 7)
;;
;; Return the size in bytes of the serialised representation of `x`,
;; or `#f` if `x` is not serialisable.
;; The `pos8` argument indicates the alignment 8-byte alignment (which
;; is necessary for unboxed vector types).
;;
(define (size/aligned x pos8)
  (cond
    ;; NOTE: order of tests on number types matters!
    [(fixnum? x)   (size-word/aligned x pos8)]
    [(flonum? x)   (size-word/aligned x pos8)]
    [(boolean? x)  (size-null/aligned x pos8)]
    [(integer? x)  (size-uvec-integer/aligned x pos8)]
;   [(rational? x) (size-pair-rational/aligned x pos8)]
;   ;; rational? NOT currently supported by pycket
;   [(complex? x)  (size-pair-complex/aligned x pos8)]
;   ;; complex? NOT currently supported by pycket

    [(pair? x) (size-pair-cons/aligned x pos8)]

    ;; NOTE: order of tests on vector types matters!
;   [(fxvector? x) (size-uvec-fxvector/aligned x pos8)]
;   ;; fxvector? NOT currently supported by pycket
    [(flvector? x) (size-uvec-flvector/aligned x pos8)]
    [(vector? x)   (size-vec-vector/aligned x pos8)]
    [(struct? x)   (size-vec-struct/aligned x pos8)]

    [(bytes? x)  (size-str-bytes/aligned  x pos8)]
    [(string? x) (size-str-string/aligned x pos8)]

    [(symbol? x) (size-sym/aligned x pos8)]

    [(hash? x) (size-dict-hash/aligned x pos8)]

    [(null? x) (size-null/aligned x pos8)]
    [(void? x) (size-null/aligned x pos8)]

    [else #f]))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Serialisation

;; (serialize x [size]) -> (and/c bytes? immutable?)
;;   x    : any/c
;;   size : non-negative-integer? = 0
;;
;; Return a bytestring containing the serialized representation of `x`;
;; the length of the output is a multiple of 8.
;; The `size` argument is an upper bound on the size of output; if `size` is
;; positive then the output is guaranteed to be at most 7 bytes longer
;; than `size`. If `size` is `0` (the default) then the required size
;; is computed by calling `serialized-size`.
;; Reaction to errors (unserialisable values, or insufficient space) is
;; the same as for `serialize-to!`.
;;
(define (serialize x [size 0])
  (define sz (if (= 0 size) (serialized-size x) size))
  (unless sz (raise 'not-serializable))
  (define buf (make-bytes (align8 sz) tag-skip))
  (serialize-to! x buf 0)
  (bytes->immutable-bytes buf))


;; (serialize-to! x buf pos) -> exact-positive-integer?
;;   x   : any/c
;;   buf : (and/c bytes? mutable?)
;;   pos : exact-nonnegative-integer?
;;
;; Serializes `x` to `buf` from position `pos` (inclusive) onwards and
;; returns the next free position in `buf`.
;; Throws an exception if `x` can't be serialised, or if there is not
;; enough space in `buf` from `pos` onwards; in both these cases the
;; contents of `buf` from `pos` onwards are undefined.
;; Note: It is assumed that `(bytes-length buf)` is a multiple of 8.
;;
(define (serialize-to! x buf pos)
  (cond
    ;; NOTE: order of tests on number types matters!
    [(fixnum? x)   (serialize-word-fixnum-to! x buf pos)]
    [(flonum? x)   (serialize-word-flonum-to! x buf pos)]
    [(boolean? x)  (serialize-null-boolean-to! x buf pos)]
    [(integer? x)  (serialize-uvec-integer-to! x buf pos)]
;   [(rational? x) (serialize-pair-rational-to! x buf pos)]
;   ;; rational? NOT currently supported by pycket
;   [(complex? x)  (serialize-pair-complex-to! x buf pos)]
;   ;; complex? NOT currently supported by pycket

    [(pair? x) (serialize-pair-cons-to! x buf pos)]

    ;; NOTE: order of tests on vector types matters!
;   [(fxvector? x) (serialize-uvec-fxvector-to! x buf pos)]
;   ;; fxvector? NOT currently supported by pycket
    [(flvector? x) (serialize-uvec-flvector-to! x buf pos)]
    [(vector? x)   (serialize-vec-vector-to! x buf pos)]
    [(struct? x)   (serialize-vec-struct-to! x buf pos)]

    [(bytes? x)  (serialize-str-bytes-to!  x buf pos)]
    [(string? x) (serialize-str-string-to! x buf pos)]

    [(symbol? x) (serialize-sym-to! x buf pos)]

    [(hash? x) (serialize-dict-hash-to! x buf pos)]

    [(null? x) (serialize-null-nil-to!  x buf pos)]
    [(void? x) (serialize-null-void-to! x buf pos)]

    [else (raise 'non-serializable-type)]))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Deserialisation

;; (deserialize buf) -> any/c
;;   buf : bytes?
;;
;; Deserializes `buf` and returns the deserialized value.
;; Reaction to errors is the same as for `deserialize-from`.
;; Note: It is assumed that `(bytes-length buf)` is a multiple of 8.
;;
(define (deserialize buf)
  (define-values (x pos) (deserialize-from buf 0))
  x)


;; (deserialize-from buf pos) -> (any/c exact-positive-integer?)
;;   buf : bytes?
;;   pos : exact-nonnegative-integer?
;;
;; Deserializes `buf` from position `pos` (inclusive) onwards and
;; returns the deserialized value and the next position in `buf`.
;; Throws an exception if deserialisation fails either because of a parsing
;; error or because of running out of input.
;; Note: It is assumed that `(bytes-length buf)` is a multiple of 8.
;;
(define (deserialize-from buf pos)
  (define tag (bytes-ref buf pos))
  (define deserialize-from/tag (vector-ref dispatch (arithmetic-shift tag -3)))
  (deserialize-from/tag buf pos tag))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Tags (of type `byte?`) used in serialised encoding;
;; lowest 3 bits are are not significant.
;; NOTE: Tag (* 8 #b10111) is currently unused.

(define tag-skip  (* 8 #b00000))  ;; skip till next 8-byte aligned position
(define tag-sym   (* 8 #b01000))  ;; symbols
(define tag-null  (* 8 #b10000))  ;; nullary data constructors (incl booleans)
(define tag-word  (* 8 #b10001))  ;; machine-word sized (64 bit) data
(define tag-pair  (* 8 #b10010))  ;; pairs (incl rational and complex numbers)
(define tag-dict  (* 8 #b10011))  ;; dictionaries (ie. hash tables)
(define tag-str   (* 8 #b10100))  ;; strings and bytestrings
(define tag-vec   (* 8 #b10101))  ;; boxed vectors (incl prefab structs)
(define tag-uvec  (* 8 #b10110))  ;; unboxed vectors (incl big integers)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Symbols
;;
;; NOTE: symbols must consist entirely of 7-bit ASCII characters;
;;       otherwise serialisation will fail (raising 'non-ASCII-symbol).
;;
;; Tag #b01nnnnnn, where
;; * nnnnnn encodes the length of the symbol (in bytes);
;;   nnnnnn = 0 means the length is read from the byte following the tag.

(define (tag-sym-len tag) (bitwise-and #b111111 tag))

;; NOTE: `len` must be between 0 and 63.
(define (make-tag-sym len) (bitwise-ior tag-sym len))


(define (size-sym/aligned x pos8)
  (define str (symbol->string x))
  (define len (string-length str))
  (and
    (symbol-interned? x) (<= len 255) (string-ascii? str)
    (if (< len 64) (+ 1 len) (+ 2 len))))


(define (serialize-sym-to! x buf pos)
  (define str (symbol->string x))
  (define len (string-length str))
  (define pos1
    (if (< len 64)
      (begin 
        (bytes-set! buf pos (make-tag-sym len))
        (+ pos 1))
      (begin
        (bytes-set! buf  pos       tag-sym) ;; Length bits set ot 0
        (bytes-set! buf  (+ pos 1) len)     ;; Length assumed <= 255
        (+ pos 2))))
  (define bstr (string->bytes/utf-8 str))
  (bytes-copy! buf pos1 bstr)
  (if (= len (bytes-length bstr))
    (+ pos1 len)
    (raise 'non-ASCII-symbol)))


(define (deserialize-sym-from/tag buf pos tag)
  (define-values (len pos1)
    (let ([n (tag-sym-len tag)])
      (if (= 0 n)
        (values (bytes-ref buf (+ pos 1)) (+ pos 2))
        (values n (+ pos 1)))))
  (values
    (string->symbol (bytes->string/utf-8 (subbytes buf pos1 (+ pos1 len))))
    (+ pos1 len)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Nullary data constructors
;;
;; Tag #b10000xcc, where
;; * cc encodes constructor:
;;   + 00 = #f
;;   + 01 = #t
;;   + 10 = '()
;;   + 11 = #<void>

(define (tag-null-#f?      tag) (= (bitwise-and #b11 tag) 0))
(define (tag-null-#t?      tag) (= (bitwise-and #b11 tag) 1))
(define (tag-null-nil?     tag) (= (bitwise-and #b11 tag) 2))
(define (tag-null-void?    tag) (= (bitwise-and #b11 tag) 3))

(define tag-null-#f                     tag-null)
(define tag-null-#t   (bitwise-ior #b01 tag-null))
(define tag-null-nil  (bitwise-ior #b10 tag-null))
(define tag-null-void (bitwise-ior #b11 tag-null))


(define (size-null/aligned x pos8) 1)


(define (serialize-null-boolean-to! x buf pos)
  (bytes-set! buf pos (if x tag-null-#t tag-null-#f))
  (+ pos 1))

(define (serialize-null-nil-to! x buf pos)
  (bytes-set! buf pos tag-null-nil)
  (+ pos 1))

(define (serialize-null-void-to! x buf pos)
  (bytes-set! buf pos tag-null-void)
  (+ pos 1))


(define (deserialize-null-from/tag buf pos tag)
  (cond
    [(tag-null-#f?   tag) (values #f     (+ pos 1))]
    [(tag-null-#t?   tag) (values #t     (+ pos 1))]
    [(tag-null-nil?  tag) (values '()    (+ pos 1))]
    [(tag-null-void? tag) (values (void) (+ pos 1))]))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Machine-word sized data (assuming 64-bit architecture)
;;
;; NOTE: data is not aligned.
;;
;; Tag #b10001xxt, where
;; * t encodes type:
;;   + 0 = fixnum?
;;   + 1 = flonum?

(define (tag-word-type tag) (bitwise-and #b1 tag))

(define tag-word-fixnum                  tag-word)
(define tag-word-flonum (bitwise-ior #b1 tag-word))


(define (size-word/aligned x pos8) 9)


(define (serialize-word-fixnum-to! x buf pos)
  (bytes-set! buf pos tag-word-fixnum)
  (bytes-copy! buf (+ pos 1) (integer->integer-bytes x 8 #t))
  (+ pos 9))

(define (serialize-word-flonum-to! x buf pos)
  (bytes-set! buf pos tag-word-flonum)
  (bytes-copy! buf (+ pos 1) (real->floating-point-bytes x 8))
  (+ pos 9))


(define (deserialize-word-from/tag buf pos tag)
  (if (= 0 (tag-word-type tag))
    (values
      (integer-bytes->integer (subbytes buf (+ pos 1) (+ pos 9)) #t)
      (+ pos 9))
    (values
      (floating-point-bytes->real (subbytes buf (+ pos 1) (+ pos 9)))
      (+ pos 9))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Pairs (incl exact-rational and complex numbers)
;;
;; NOTE: data is not aligned.
;;
;; Tag #b10010xtt, where
;; * tt encodes the type
;;   + 00 = pair?
;;   + 10 = rational?
;;   + 11 = complex?

(define (tag-pair-cons?     tag) (= (bitwise-and #b11 tag) 0))
(define (tag-pair-rational? tag) (= (bitwise-and #b11 tag) 2))
(define (tag-pair-complex?  tag) (= (bitwise-and #b11 tag) 3))

(define tag-pair-cons                       tag-pair)
(define tag-pair-rational (bitwise-ior #b10 tag-pair))
(define tag-pair-complex  (bitwise-ior #b11 tag-pair))


(define (size-pair/aligned x pos8 fst snd)
  (let ([n1 (size/aligned (fst x) (mod8 (+ pos8 1)))])
    (and n1
      (let ([n2 (size/aligned (snd x) (mod8 (+ pos8 1 n1)))])
        (and n2
          (+ 1 n1 n2))))))

(define (size-pair-cons/aligned x pos8)
  (size-pair/aligned x pos8 car cdr))

(define (size-pair-rational/aligned x pos8)
  (size-pair/aligned x pos8 numerator denominator))

(define (size-pair-complex/aligned x pos8)
  (size-pair/aligned x pos8 real-part imag-part))


(define (serialize-pair-to! x buf pos fst snd tag)
  (bytes-set! buf pos tag)
  (serialize-to! (snd x) buf (serialize-to! (fst x) buf (+ pos 1))))

(define (serialize-pair-cons-to! x buf pos)
  (serialize-pair-to! x buf pos car cdr tag-pair-cons))

(define (serialize-pair-rational-to! x buf pos)
  (serialize-pair-to! x buf pos numerator denominator tag-pair-rational))

(define (serialize-pair-complex-to! x buf pos)
  (serialize-pair-to! x buf pos real-part imag-part tag-pair-complex))


(define (deserialize-pair-from/tag buf pos tag)
  (define-values (x1 pos1) (deserialize-from buf (+ pos 1)))
  (define-values (x2 pos2) (deserialize-from buf pos1))
  (cond
    [(tag-pair-cons?     tag) (values (cons x1 x2) pos2)]
    [(tag-pair-rational? tag) (values (/ x1 x2) pos2)]
    [(tag-pair-complex?  tag) (values (make-rectangular x1 x2) pos2)]))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Dictionaries (ie hash tables)
;;
;; NOTE: data is not aligned.
;;
;; Tag #b10011xtt, where
;; * tt encodes the type
;;   + 00 = hash-eq?
;;   + 01 = hash-eqv?
;;   + 10 = hash-equal?

(define (tag-dict-hash-eq?    tag) (= (bitwise-and #b11 tag) 0))
(define (tag-dict-hash-eqv?   tag) (= (bitwise-and #b11 tag) 1))
(define (tag-dict-hash-equal? tag) (= (bitwise-and #b11 tag) 2))

(define tag-dict-hash-eq                      tag-dict)
(define tag-dict-hash-eqv   (bitwise-ior #b01 tag-dict))
(define tag-dict-hash-equal (bitwise-ior #b10 tag-dict))


(define (size-dict-hash/aligned x pos8)
  (define n (size/aligned (hash->list x) (mod8 (+ pos8 1))))
  (and n (+ n 1)))


(define (serialize-dict-hash-to! x buf pos)
  (define tag (cond
                [(hash-eq?    x) tag-dict-hash-eq]
                [(hash-eqv?   x) tag-dict-hash-eqv]
                [(hash-equal? x) tag-dict-hash-equal]))
  (bytes-set! buf pos tag)
  (serialize-to! (hash->list x) buf (+ pos 1)))


(define (deserialize-dict-from/tag buf pos tag)
  (define-values (assocs pos1) (deserialize-from buf (+ pos 1)))
  (cond
    [(tag-dict-hash-eq?    tag) (values (make-immutable-hasheq assocs)  pos1)]
    [(tag-dict-hash-eqv?   tag) (values (make-immutable-hasheqv assocs) pos1)]
    [(tag-dict-hash-equal? tag) (values (make-immutable-hash assocs)    pos1)]))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Strings and bytestrings
;;
;; NOTE: strings must consist entirely of 7-bit ASCII characters;
;;       otherwise serialisation will fail (raising 'non-ASCII-string).
;;       To serialise other strings, encode first as UTF8 bytestrings.
;;
;; Tag #b10100sxt, where
;; * s encodes size of length information (see boxed vectors), and
;; * t encodes the type
;;   + 0 = bytes?
;;   + 1 = string?

(define (tag-str-type tag) (bitwise-and #b1 tag))

(define tag-str-bytes                   tag-str)
(define tag-str-string (bitwise-ior #b1 tag-str))


(define (size-str-bytes/aligned x pos8)
  (define n (bytes-length x))
  (if (<= n 255) (+ 2 n) (+ 9 n)))

(define (size-str-string/aligned x pos8)
  (define n (string-length x))
  (and
    (string-ascii? x)
    (if (<= n 255) (+ 2 n) (+ 9 n))))


(define (serialize-str-bytes-to! x buf pos)
  (define len (bytes-length x))
  (define pos1 (write-tag+len-to! tag-str-bytes len buf pos))
  (bytes-copy! buf pos1 x)
  (+ pos1 len))

(define (serialize-str-string-to! x buf pos)
  (define len (string-length x))
  (define pos1 (write-tag+len-to! tag-str-string len buf pos))
  (define bstr (string->bytes/utf-8 x))
  (bytes-copy! buf pos1 bstr)
  (if (= len (bytes-length bstr))
    (+ pos1 len)
    (raise 'non-ASCII-string)))


(define (deserialize-str-from/tag buf pos tag)
  (define-values (len pos1) (read-len-from/tag buf (+ pos 1) tag))
  (if (= 0 (tag-str-type tag))
    (values
      (subbytes buf pos1 (+ pos1 len))
      (+ pos1 len))
    (values
      (bytes->string/utf-8 (subbytes buf pos1 (+ pos1 len)))
      (+ pos1 len))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Boxed vectors and prefab structs;
;; also includes code for reading and writing length of vectors, strings, etc
;;
;; NOTE: data is not aligned.
;;
;; Tag #b10101sxt, where
;; * s encodes size of length information
;;   + 0 = 1 byte (0 <= length <= 255)
;;   + 1 = 8 byte
;; * t encodes the type
;;   + 0 = vector?
;;   + 1 = struct?  (ONLY PREFAB STRUCTS)

(define (tag-vec-type tag) (bitwise-and #b1 tag))

(define tag-vec-vector                  tag-vec)
(define tag-vec-struct (bitwise-ior #b1 tag-vec))

(define (tag-len-size tag) (bitwise-and #b100 tag))

(define (set-tag-len-size-1 tag)                    tag)
(define (set-tag-len-size-8 tag) (bitwise-ior #b100 tag))


(define (size-vec-vector/aligned x pos8)
  (define n (vector-length x))
  (for/fold ([sum (if (<= n 255) 2 9)]) ([x_i x])
    (and sum
      (let ([s_i (size/aligned x_i (mod8 (+ pos8 sum)))])
        (and s_i (+ sum s_i))))))

(define (size-vec-struct/aligned x pos8)
  (define v (prefab-struct->vector x))
  (and v (size-vec-vector/aligned v pos8)))


(define (serialize-vec-to! x buf pos tag)
  (define len (vector-length x))
  (define pos1 (write-tag+len-to! tag len buf pos))
  (for/fold ([pos_i pos1]) ([x_i x])
    (serialize-to! x_i buf pos_i)))

(define (serialize-vec-vector-to! x buf pos)
  (serialize-vec-to! x buf pos tag-vec-vector))

(define (serialize-vec-struct-to! x buf pos)
  (define v (prefab-struct->vector x))
  (unless v (raise 'non-serializable-struct-type))
  (serialize-vec-to! v buf pos tag-vec-struct))


(define (deserialize-vec-from/tag buf pos tag)
  (define-values (n pos1) (read-len-from/tag buf (+ pos 1) tag))
  (define-values (v pos2) (read-vec-from buf pos1 n))
  (if (= 0 (tag-vec-type tag))
    (values (vector->immutable-vector v) pos2)
    (values (vector->prefab-struct    v) pos2)))


;; Reads a vector of `n` items from `buf` from `pos` onwards;
;; returns the (mutable) vector and next position.
(define (read-vec-from buf pos n)
  (define v (make-vector n))
  (define pos_n (for/fold ([pos_i pos]) ([i (in-range n)])
                  (define-values (v_i pos_i+1) (deserialize-from buf pos_i))
                  (vector-set! v i v_i)
                  pos_i+1))
  (values v pos_n))


;; Writes `tag` and `len` to `buf` from position `pos`; returns next position.
(define (write-tag+len-to! tag len buf pos)
  (if (<= len 255)
    (begin
      (bytes-set! buf pos       (set-tag-len-size-1 tag))
      (bytes-set! buf (+ pos 1) len)
      (+ pos 2))
    (begin
      (bytes-set!  buf pos       (set-tag-len-size-8 tag))
      (bytes-copy! buf (+ pos 1) (integer->integer-bytes len 8 #f))
      (+ pos 9))))


;; Reads length (1 or 8 bytes according to `tag`) from `buf` from `pos` onwards;
;; returns length and next position.
(define (read-len-from/tag buf pos tag)
  (if (= 0 (tag-len-size tag))
    (values
      (bytes-ref buf pos)
      (+ pos 1))
    (values
      (integer-bytes->integer (subbytes buf pos (+ pos 8)) #f)
      (+ pos 8))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Unboxed vectors (incl integers too big for fixnum?)
;;
;; NOTE: payload data is word-aligned (but tag and length are not).
;;
;; Tag #b10110stt, where
;; * s encodes size of length information (see boxed vectors), and
;; * tt encodes the type
;;   + 00 = positive integer?
;;   + 01 = negative integer?
;;   + 10 = fxvector?
;;   + 11 = flvector?

(define (tag-uvec-integer?  tag) (< (bitwise-and #b11 tag) 2))
(define (tag-uvec-fxvector? tag) (= (bitwise-and #b11 tag) 2))
(define (tag-uvec-flvector? tag) (= (bitwise-and #b11 tag) 3))

(define (tag-uvec-integer-neg tag) (bitwise-and #b1 tag))

;; NOTE: `neg` must be 0 or 1.
(define (make-tag-uvec-integer neg) (bitwise-ior tag-uvec neg))

(define tag-uvec-fxvector (bitwise-ior #b10 tag-uvec))
(define tag-uvec-flvector (bitwise-ior #b11 tag-uvec))


(define (size-uvec/aligned x pos8 length)
  (define n (length x))
  (define header (- (align8 (+ pos8 (if (<= n 255) 2 9))) pos8))
  (+ header (* 8 n)))

(define (size-uvec-fxvector/aligned x pos8)
  (size-uvec/aligned x pos8 fxvector-length))

(define (size-uvec-flvector/aligned x pos8)
  (size-uvec/aligned x pos8 flvector-length))

(define (size-uvec-integer/aligned x pos8)
  (size-uvec/aligned x pos8 integer-list-length))


(define (serialize-uvec-to! x buf pos length encode tag)
  (define len (length x))
  (define pos1 (align8 (write-tag+len-to! tag len buf pos)))
  (for ([i (in-range len)] [x_i x])
    (bytes-copy! buf (+ pos1 (* 8 i)) (encode x_i)))
  (+ pos1 (* 8 len)))

(define (serialize-uvec-fxvector-to! x buf pos)
  (define (encode i) (integer->integer-bytes i 8 #t))
  (serialize-uvec-to! x buf pos fxvector-length encode tag-uvec-fxvector))

(define (serialize-uvec-flvector-to! x buf pos)
  (define (encode r) (real->floating-point-bytes r 8))
  (serialize-uvec-to! x buf pos flvector-length encode tag-uvec-flvector))

(define (serialize-uvec-integer-to! x buf pos)
  (define tag (make-tag-uvec-integer (if (< x 0) 1 0)))
  (define-values (len list) (integer->integer-list x))
  (define pos1 (align8 (write-tag+len-to! tag len buf pos)))
  (write-integer-list-to! list buf pos1))
  

(define (deserialize-uvec-from/tag buf pos tag)
  (define-values (len pos0) (read-len-from/tag buf (+ pos 1) tag))
  (define pos1 (align8 pos0))
  (cond
    [(tag-uvec-integer? tag)
       (let* ([n (read-integer-from buf pos1 len)]
              [x (if (= 0 (tag-uvec-integer-neg tag)) n (- n))])
         (values x (+ pos1 (* 8 len))))]
    [(tag-uvec-fxvector? tag)
       (let ([x (make-fxvector len)])
         (for ([i (in-range len)])
           (define pos_i (+ pos1 (* 8 i)))
           (define x_i (integer-bytes->integer
                         (subbytes buf pos_i (+ pos_i 8)) #t))
           (fxvector-set! x i x_i))
         (values x (+ pos1 (* 8 len))))]    ;; NOTE: `x` is NOT immutable!
    [(tag-uvec-flvector? tag)
       (let ([x (make-flvector len)])
         (for ([i (in-range len)])
           (define pos_i (+ pos1 (* 8 i)))
           (define x_i (floating-point-bytes->real
                         (subbytes buf pos_i (+ pos_i 8))))
           (flvector-set! x i x_i))
         (values x (+ pos1 (* 8 len))))]))  ;; NOTE: `x` is NOT immutable!


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Deserialisation dispatcher
;;

;; Deserializer for tagless words (fixnums with most significant octet first).
(define (deserialize-word/tagless-from/tag buf pos tag)
  (raise 'big-endian-fixnum-not-implemented))

;; Deserializer for skip tags, skipping to next 8-byte aligned position.
(define (deserialize-skip-from/tag buf pos tag)
  (if (= 0 (modulo pos 8))
    (deserialize-word/tagless-from/tag buf pos tag)
    (deserialize-from buf (align8 pos))))

(define (deserialize-unimpl-from/tag buf pos tag)
  (raise 'not-implemented))

;; Dispatch vector used by deserializer; indexed by highest 5 tag bits.
(define dispatch
  (vector
    deserialize-skip-from/tag           ;; #b00000
    deserialize-word/tagless-from/tag   ;; #b00001
    deserialize-word/tagless-from/tag   ;; #b00010
    deserialize-word/tagless-from/tag   ;; #b00011
    deserialize-word/tagless-from/tag   ;; #b00100
    deserialize-word/tagless-from/tag   ;; #b00101
    deserialize-word/tagless-from/tag   ;; #b00110
    deserialize-word/tagless-from/tag   ;; #b00111

    deserialize-sym-from/tag            ;; #b01000
    deserialize-sym-from/tag            ;; #b01001
    deserialize-sym-from/tag            ;; #b01010
    deserialize-sym-from/tag            ;; #b01011
    deserialize-sym-from/tag            ;; #b01100
    deserialize-sym-from/tag            ;; #b01101
    deserialize-sym-from/tag            ;; #b01110
    deserialize-sym-from/tag            ;; #b01111

    deserialize-null-from/tag           ;; #b10000
    deserialize-word-from/tag           ;; #b10001
    deserialize-pair-from/tag           ;; #b10010
    deserialize-dict-from/tag           ;; #b10011
    deserialize-str-from/tag            ;; #b10100
    deserialize-vec-from/tag            ;; #b10101
    deserialize-uvec-from/tag           ;; #b10110
    deserialize-unimpl-from/tag         ;; #b10111

    deserialize-word/tagless-from/tag   ;; #b11000
    deserialize-word/tagless-from/tag   ;; #b11001
    deserialize-word/tagless-from/tag   ;; #b11010
    deserialize-word/tagless-from/tag   ;; #b11011
    deserialize-word/tagless-from/tag   ;; #b11100
    deserialize-word/tagless-from/tag   ;; #b11101
    deserialize-word/tagless-from/tag   ;; #b11110
    deserialize-word/tagless-from/tag)) ;; #b11111


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Conversion between nonnegative exact integers and lists of digits

;; Bit-width of digits (bounded by 64 so that a digit fits into 8 bytes)
(define integer-digit-bits 63)
(define integer-digit-mask (sub1 (expt 2 integer-digit-bits)))


;; Counts the number of digits returned by `integer->integer-list`.
(define (integer-list-length x)
  (define (count-digits n len)
    (if (= 0 n)
      len
      (let ([n1 (arithmetic-shift n (- integer-digit-bits))])
        (count-digits n1 (+ len 1)))))
  (count-digits (abs x) 0))


;; Returns a list of digits corresponding to the absolute value of integer `x`
;; (starting with the most significant digit) and the length of that list.
(define (integer->integer-list x)
  (define (gather-digits n len list)
    (if (= 0 n)
      (values len list)
      (let ([digit (bitwise-and integer-digit-mask n)]
            [n1    (arithmetic-shift n (- integer-digit-bits))])
        (gather-digits n1 (+ len 1) (cons digit list)))))
  (gather-digits (abs x) 0 '()))

;; Converts a list of digits (starting with the most significant digit)
;; into a nonnegative integer.
(define (integer-list->integer digits)
  (for/fold ([n 0]) ([digit digits])
    (bitwise-ior digit (arithmetic-shift n integer-digit-bits))))


;; Writes the list of integers `digits` (8-byte wide) to `buf`
;; from position `pos` onwards, and returns the next position.
(define (write-integer-list-to! digits buf pos)
  (for/fold ([current-pos pos]) ([digit digits])
    (bytes-copy! buf current-pos (integer->integer-bytes digit 8 #f))
    (+ current-pos 8)))


;; Writes (the absolute value of) integer `n` to `buf` from position `pos`,
;; encoded as list of 8-byte digits, starting with the most significant digit.
;; Returns the number of digits written.
(define (write-integer-to! n buf pos)
  (define-values (len digits) (integer->integer-list n))
  (write-integer-list-to! digits buf pos)
  len)

;; Reads `len` 8-byte digits (starting with most significant digit)
;; from `buf` from position `pos` onwards, and returns the corresponding
;; nonnegative integer.
(define (read-integer-from buf pos len)
  (for/fold ([n 0]) ([i (in-range len)])
    (define pos_i (+ pos (* 8 i)))
    (define d_i   (integer-bytes->integer (subbytes buf pos_i (+ pos_i 8)) #f))
    (bitwise-ior d_i (arithmetic-shift n integer-digit-bits))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Functions from racket libraries that are missing in pycket

(define (bytes->immutable-bytes bstr) bstr)
;; NOTE: Dummy implementation; result is NOT immutable.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Misc auxiliariy functions

(define (mod8 n)
  (modulo n 8))


;; Adjust nonnegative integer `n` upwards to the nearest multiple of 8.
(define (align8 n)
  (* (quotient (+ n 7) 8) 8))


;; Returns `#t` iff string `str` consists exclusively of 7-bit ASCII characters.
;; WARNING: This test is inefficient for long strings.
(define (string-ascii? str)
  (for/and ([c str]) (< (char->integer c) 128)))


;; Returns a vector corresponding to struct `st` if `st` is prefab; `#f` else.
(define (prefab-struct->vector st)
  (define key (prefab-struct-key st))
  (and key
    (let ([v (struct->vector st)])
      (vector-set! v 0 key)
      v)))


;; Inverse of `prefab-struct->vector`.
(define (vector->prefab-struct v)
  (apply make-prefab-struct (vector->list v)))
